package batoh;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;
import java.util.Vector;
import java.util.concurrent.Semaphore;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Implementovat hrubou silou, hladovy algoritmus, spocitat chybu
 *
 * @author karry
 */
public class Batoh {

  private static Logger logger = Logger.getLogger(Batoh.class.getName());
  private static boolean debug = false;
  private volatile int threadsNum = 0;
  public static final boolean THREAD_LIMIT = true;
  public static final int DEFAULT_MAX_THREADS = 20;
  public static int MAX_THREADS = DEFAULT_MAX_THREADS;

  /**
   * @param args the command line arguments
   */
  public static void main(String[] args) throws InterruptedException {
    float nosnost = 0;
    if (args.length < 1) {
      logger.log(Level.SEVERE, "Prvnim argumentem musi byt nosnost batohu!");
      System.exit(-1);
    } else {
      try {
        nosnost = Float.parseFloat(args[0]);
      } catch (NumberFormatException ex) {
        logger.log(Level.SEVERE, "Prvnim argument njede prevest na float!");
        System.exit(-1);
      }
    }

    if (args.length >= 2){
      try{
        MAX_THREADS = Integer.parseInt(args[1]);
        if (MAX_THREADS < 0)
          throw new IllegalArgumentException("Druhy argument (maximalni pocet vlaken) nemuze byt zporny!");
      } catch (NumberFormatException ex) {
        logger.log(Level.SEVERE, "Prvnim argument njede prevest na float!");
        System.exit(-1);
      }
    }

    Batoh batoh = new Batoh(nosnost);
    Properties prop = new Properties();

    try {
      prop.load(new InputStreamReader(new BufferedInputStream(new FileInputStream("predmety.ini"))));
      batoh.nacti(prop);

      System.out.println("sezman veci:");
      Batoh.vypis(batoh.getPredmety());

      System.out.println("\nhladovy algoritmus:");
      Vector<Predmet> hladovyReseni = batoh.hladovyAlg();
      Batoh.vypis(hladovyReseni);
      System.out.println("celkova cena: " + Batoh.cena(hladovyReseni) + " " +
              "(vaha/nosnost " + Batoh.vaha(hladovyReseni) + "/" + batoh.getNosnost() + ")");

      System.out.println("\nalgoritmus 'hrubou silou':");
      long startTime = System.currentTimeMillis();
      Vector<Predmet> brutalReseni = batoh.brutalForceAlg();
      System.out.println("");
      Batoh.vypis(brutalReseni);
      System.out.println("celkova cena: " + Batoh.cena(brutalReseni) + " " +
              "(vaha/nosnost " + Batoh.vaha(brutalReseni) + "/" + batoh.getNosnost() + ")");
      System.out.println("Algoritmus trval " + (System.currentTimeMillis() - startTime) + "ms");

    } catch (IOException ex) {
      logger.log(Level.SEVERE, ex.getMessage(), ex);
      System.exit(-1);
    } catch (RuntimeException ex) {
      logger.log(Level.SEVERE, ex.getMessage(), ex);
      System.exit(-1);
    }
  }

  private static float vaha(Vector<Predmet> batoh) {
    float result = 0;
    for (Predmet predmet : batoh) {
      result += predmet.getVaha();
    }
    return result;
  }

  private static float cena(Vector<Predmet> batoh) {
    float result = 0;
    for (Predmet predmet : batoh) {
      result += predmet.getCena();
    }
    return result;
  }

  private static String tab(int depth) {
    String result = "";
    for (; depth > 0; depth--) {
      result += " ";
    }

    return result;
  }
  private final float nosnost;
  private Vector<Predmet> predmety = new Vector<Predmet>();

  public Batoh(float nosnost) {
    if (nosnost < 0) {
      throw new RuntimeException("Nosnost nemuze byt zaporna!");
    }
    this.nosnost = nosnost;
  }

  public Vector<Predmet> getPredmety() {
    return predmety;
  }

  private void nacti(Properties prop) throws NumberFormatException {
    for (Object key : prop.keySet()) {
      String name = (String) key;
      String val = (String) prop.get(key);
      String[] arr = val.split(";");
      if (arr.length != 2) {
        throw new RuntimeException("spatny format vstupnich dat...");
      }

      predmety.add(new Predmet(Float.parseFloat(arr[0]), Float.parseFloat(arr[1]), name));
    }
  }

  private static void vypis(Vector<Predmet> predmety) {
    for (Predmet predmet : predmety) {
      System.out.println(predmet);
    }
  }

  private Vector<Predmet> hladovyAlg() {
    Vector<Predmet> tmpPredmety = new Vector<Predmet>(predmety);

    Vector<Predmet> batoh = new Vector<Predmet>();
    while (!tmpPredmety.isEmpty()) {
      float nejlepsi = Float.MIN_VALUE;
      Predmet best = null;
      for (Predmet p : tmpPredmety) {
        if (p.getVaha() == 0) {
          best = p;
          break;
        }
        if (nejlepsi < p.getPomer()) {
          nejlepsi = p.getPomer();
          best = p;
        }
      }
      if (best == null) {
        throw new RuntimeException("neco je spatne... selhal vypocest nejlepsi veci");
      }

      if ((Batoh.vaha(batoh) + best.getVaha()) <= nosnost) {
        batoh.add(best);
      }

      tmpPredmety.remove(best);
    }

    return batoh;
  }

  private Vector<Predmet> brutalForceAlg() throws InterruptedException {
    //return Batoh.brutalRecursion(predmety, new Vector<Predmet>(), nosnost, 0);
    RecursionRunnable r = new RecursionRunnable(predmety, new Vector<Predmet>(), nosnost, 0);
    (new Thread(r)).start();
    return r.getResult();
  }

  private float getNosnost() {
    return nosnost;
  }

  class RecursionRunnable implements Runnable {

    private final Vector<Predmet> veci;
    private final Vector<Predmet> batoh;
    volatile private Vector<Predmet> result;
    private final int depth;
    private Semaphore runSemaphore = new Semaphore(0);
    private InterruptedException ex = null;

    RecursionRunnable(Vector<Predmet> veci, Vector<Predmet> batoh, float nosnost, int depth) {
      this.veci = veci;
      this.batoh = batoh;
      this.depth = depth;
    }

    public void run(){
      try {
        process();
      } catch (InterruptedException ex) {
        this.ex = ex;
        Logger.getLogger(Batoh.class.getName()).log(Level.SEVERE, null, ex);
      }
      threadsNum --;
    }

    public void process() throws InterruptedException{

      if (veci.isEmpty()) {
        this.result = batoh;
      }
      Vector<RecursionRunnable> threads = new Vector<RecursionRunnable>();

      Vector<Predmet> best = batoh;
      for (Predmet p : veci) {
        if (Batoh.vaha(batoh) + p.getVaha() > nosnost) {
          continue;
        }

        Vector<Predmet> tmp = new Vector<Predmet>(veci);
        Vector<Predmet> tmpBatoh = new Vector<Predmet>(batoh);
        tmp.remove(p);
        tmpBatoh.add(p);

        RecursionRunnable r = new RecursionRunnable(tmp, tmpBatoh, nosnost, depth + 1);
        threads.add(r);
        if (threadsNum >= MAX_THREADS && THREAD_LIMIT){
          r.process();
        }else{
          threadsNum ++;
          (new Thread(r)).start();
        }
      }

      for (RecursionRunnable r: threads){
        //Vector<Predmet> test = brutalRecursion(tmp, tmpBatoh, nosnost, depth + 1);
        Vector<Predmet> test = r.getResult();
        if (cena(test) > cena(best)) {
          best = test;
        }
      }
      if (debug) {
        System.out.println(depth + "> " + Batoh.tab(depth) + cena(best));
      }
      this.result = best;
      runSemaphore.release();
    }

    public Vector<Predmet> getResult() throws InterruptedException{
      if (ex != null)
        throw ex;
      runSemaphore.acquire();
      return this.result;
    }
  }
}
